{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%pylab inline\n",
    "import torch\n",
    "import sys, os\n",
    "import pystk\n",
    "import ray\n",
    "device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')\n",
    "print('device = ', device)\n",
    "ray.init(logging_level=50)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "@ray.remote\n",
    "class Rollout:\n",
    "    def __init__(self, screen_width, screen_height, hd=True, track='lighthouse', render=True):\n",
    "        # Init supertuxkart\n",
    "        if not render:\n",
    "            config = pystk.GraphicsConfig.none()\n",
    "        elif hd:\n",
    "            config = pystk.GraphicsConfig.hd()\n",
    "        else:\n",
    "            config = pystk.GraphicsConfig.ld()\n",
    "        config.screen_width = screen_width\n",
    "        config.screen_height = screen_height\n",
    "        pystk.init(config)\n",
    "        \n",
    "        self.render = render\n",
    "        race_config = pystk.RaceConfig(track=track)\n",
    "        self.race = pystk.Race(race_config)\n",
    "        self.race.start()\n",
    "    \n",
    "    def __call__(self, agent, n_steps=200):\n",
    "        torch.set_num_threads(1)\n",
    "        self.race.restart()\n",
    "        self.race.step()\n",
    "        data = []\n",
    "        track_info = pystk.Track()\n",
    "        track_info.update()\n",
    "\n",
    "        for i in range(n_steps):\n",
    "            world_info = pystk.WorldState()\n",
    "            world_info.update()\n",
    "\n",
    "            # Gather world information\n",
    "            kart_info = world_info.players[0].kart\n",
    "\n",
    "            agent_data = {'track_info': track_info, 'kart_info': kart_info}\n",
    "            if self.render:\n",
    "                agent_data['image'] = np.array(self.race.render_data[0].image)\n",
    "\n",
    "            # Act\n",
    "            action = agent(**agent_data)\n",
    "            agent_data['action'] = action\n",
    "\n",
    "            # Take a step in the simulation\n",
    "            self.race.step(action)\n",
    "\n",
    "            # Save all the relevant data\n",
    "            data.append(agent_data)\n",
    "        return data\n",
    "\n",
    "def show_video(frames, fps=30):\n",
    "    import imageio\n",
    "    from IPython.display import Video, display \n",
    "    \n",
    "    imageio.mimwrite('/tmp/test.mp4', frames, fps=fps, bitrate=10000000)\n",
    "    display(Video('/tmp/test.mp4', width=800, height=600, embed=True))\n",
    "\n",
    "viz_rollout = Rollout.remote(100, 75)\n",
    "def show_agent(agent, n_steps=600):\n",
    "    data = ray.get(viz_rollout.__call__.remote(agent, n_steps=n_steps))\n",
    "    show_video([d['image'] for d in data])\n",
    "\n",
    "rollouts = [Rollout.remote(100, 75) for i in range(10)]\n",
    "def rollout_many(many_agents, **kwargs):\n",
    "    ray_data = []\n",
    "    for i, agent in enumerate(many_agents):\n",
    "         ray_data.append( rollouts[i % len(rollouts)].__call__.remote(agent, **kwargs) )\n",
    "    return ray.get(ray_data)\n",
    "\n",
    "def dummy_agent(**kwargs):\n",
    "    action = pystk.Action()\n",
    "    action.acceleration = 1\n",
    "    return action\n",
    "\n",
    "# Let's load a fancy auto-pilot. You'll write one yourself in your homework.\n",
    "from _auto_pilot import auto_pilot\n",
    "\n",
    "show_agent(auto_pilot)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from torchvision.transforms import functional as TF\n",
    "\n",
    "class ActionNet(torch.nn.Module):\n",
    "    def __init__(self):\n",
    "        super().__init__()\n",
    "        self.network = torch.nn.Sequential(\n",
    "            torch.nn.BatchNorm2d(3),\n",
    "            torch.nn.Conv2d(3, 16, 5, stride=2),\n",
    "            torch.nn.ReLU(),\n",
    "            torch.nn.Conv2d(16, 32, 5, stride=2),\n",
    "            torch.nn.ReLU(),\n",
    "            torch.nn.Conv2d(32, 32, 5, stride=2),\n",
    "            torch.nn.ReLU(),\n",
    "            torch.nn.Conv2d(32, 32, 5, stride=2),\n",
    "            torch.nn.ReLU()\n",
    "        )\n",
    "        self.classifier = torch.nn.Linear(32, 2)\n",
    "    \n",
    "    def forward(self, x):\n",
    "        f = self.network(x)\n",
    "        return self.classifier(f.mean(dim=(2,3)))\n",
    "\n",
    "class Actor:\n",
    "    def __init__(self, action_net):\n",
    "        self.action_net = action_net.cpu().eval()\n",
    "    \n",
    "    def __call__(self, image, **kwargs):\n",
    "        output = self.action_net(TF.to_tensor(image)[None])[0]\n",
    "\n",
    "        action = pystk.Action()\n",
    "        action.acceleration = 1\n",
    "        action.steer = output[0]\n",
    "        action.drift = output[1] > 0.5\n",
    "        return action\n",
    "\n",
    "\n",
    "def noisy_actor(actor, noise_std=5):\n",
    "    def act(**kwargs):\n",
    "        action = actor(**kwargs)\n",
    "        action.steer += np.random.normal(0, noise_std)\n",
    "        return action\n",
    "    return act\n",
    "\n",
    "\n",
    "action_net = ActionNet()\n",
    "actor = Actor(action_net)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%load_ext tensorboard\n",
    "import tempfile\n",
    "log_dir = 'log_04'\n",
    "%tensorboard --logdir {log_dir} --reload_interval 1"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "scrolled": false
   },
   "outputs": [],
   "source": [
    "import torch.utils.tensorboard as tb\n",
    "from datetime import datetime\n",
    "\n",
    "n_epochs = 10\n",
    "batch_size = 128\n",
    "n_trajectories = 10\n",
    "\n",
    "# Create the network\n",
    "action_net = ActionNet().to(device)\n",
    "\n",
    "# Create the optimizer\n",
    "optimizer = torch.optim.Adam(action_net.parameters())\n",
    "\n",
    "# Create the loss\n",
    "loss = torch.nn.MSELoss()\n",
    "\n",
    "# Collect the data\n",
    "train_data = []\n",
    "for data in rollout_many([auto_pilot]*n_trajectories):\n",
    "    train_data.extend(data)\n",
    "    \n",
    "# Upload to the GPU\n",
    "train_images = torch.stack([torch.as_tensor(d['image']) for d in train_data]).permute(0,3,1,2).to(device).float()/255.\n",
    "train_labels = torch.stack([torch.as_tensor((d['action'].steer, d['action'].drift)) for d in train_data]).to(device).float()\n",
    "\n",
    "# Start training\n",
    "global_step = 0\n",
    "action_net.train().to(device)\n",
    "logger = tb.SummaryWriter(log_dir+'/'+str(datetime.now()), flush_secs=1)\n",
    "\n",
    "for epoch in range(n_epochs):\n",
    "    for iteration in range(0, len(train_data), batch_size):\n",
    "        batch_ids = torch.randint(0, len(train_data), (batch_size,), device=device)\n",
    "        batch_images = train_images[batch_ids]\n",
    "        batch_labels = train_labels[batch_ids]\n",
    "        o = action_net(batch_images)\n",
    "        loss_val = loss(o, batch_labels)\n",
    "\n",
    "        logger.add_scalar('train/loss', loss_val, global_step)\n",
    "        global_step += 1\n",
    "\n",
    "        optimizer.zero_grad()\n",
    "        loss_val.backward()\n",
    "        optimizer.step()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "show_agent(Actor(action_net), n_steps=600)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Collect the data\n",
    "for data in rollout_many([Actor(action_net)]*n_trajectories):\n",
    "    train_data.extend(data)\n",
    "# for data in rollout_many([noisy_actor(auto_pilot, noise_std=i//2) for i in range(n_trajectories)]):\n",
    "#     train_data.extend(data)\n",
    "\n",
    "# Supervise using the auto-pilot actions\n",
    "for d in train_data:\n",
    "    d['action'] = auto_pilot(**d)\n",
    "\n",
    "# Upload to the GPU\n",
    "train_images = torch.stack([torch.as_tensor(d['image']) for d in train_data]).permute(0,3,1,2).to(device).float()/255.\n",
    "train_labels = torch.stack([torch.as_tensor((d['action'].steer, d['action'].drift)) for d in train_data]).to(device).float()\n",
    "\n",
    "# Start training\n",
    "logger = tb.SummaryWriter(log_dir+'/'+str(datetime.now()), flush_secs=1)\n",
    "global_step = 0\n",
    "action_net.train().to(device)\n",
    "                                \n",
    "for epoch in range(n_epochs):\n",
    "    for iteration in range(0, len(train_data), batch_size):\n",
    "        batch_ids = torch.randint(0, len(train_data), (batch_size,), device=device)\n",
    "        batch_images = train_images[batch_ids]\n",
    "        batch_labels = train_labels[batch_ids]\n",
    "        o = action_net(batch_images)\n",
    "        loss_val = loss(o, batch_labels)\n",
    "\n",
    "        logger.add_scalar('train/loss', loss_val, global_step)\n",
    "        global_step += 1\n",
    "\n",
    "        optimizer.zero_grad()\n",
    "        loss_val.backward()\n",
    "        optimizer.step()\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "show_agent(Actor(action_net), n_steps=600)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.10"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}